# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urwid

from os.path import commonprefix


def _get_word_before_pos(p_text, p_pos):
    is_first_word = False
    text = p_text
    pos = p_pos
    start = text.rfind(' ', 0, pos) + 1

    if pos == 0 or text.lstrip().rfind(' ', 0, pos) == -1:
        is_first_word = True

    return (text[start:pos], is_first_word, start)


class CommandLineWidget(urwid.Edit):
    def __init__(self, p_completer, *args, **kwargs):

        self.history = []
        self.history_pos = None
        # temporary history storage for edits before cmd execution
        self.history_tmp = []
        self.completer = p_completer
        self.start = None  # text before completion candidate
        self.end = None    # text after completion candidate

        super().__init__(*args, **kwargs)
        urwid.register_signal(CommandLineWidget, ['blur', 'execute_command', 'show_completions'])

    def clear(self):
        self.set_edit_text("")

    def _blur(self):
        self.clear()
        urwid.emit_signal(self, 'blur')

    def _emit_command(self):
        if len(self.edit_text) > 0:
            urwid.emit_signal(self, 'execute_command', self.edit_text)
            self._save_to_history()
            self.clear()

    def _save_to_history(self):
        if len(self.edit_text) > 0:
            self.history.append(self.edit_text)

        self.history_pos = len(self.history)
        self.history_tmp = self.history[:] # sync temporary storage with real history
        self.history_tmp.append('')

    def _history_move(self, p_step):
        """
        Changes current value of the command-line to the value obtained from
        history_tmp list with index calculated by addition of p_step to the
        current position in the command history (history_pos attribute).

        Also saves value of the command-line (before changing it) to history_tmp
        for potential later access.
        """
        if len(self.history) > 0:
            # don't pollute real history - use temporary storage
            self.history_tmp[self.history_pos] = self.edit_text
            self.history_pos = self.history_pos + p_step
            self.set_edit_text(self.history_tmp[self.history_pos])

    def _history_next(self):
        if self.history_pos != len(self.history):
            self._history_move(1)

    def _history_back(self):
        if self.history_pos != 0:
            self._history_move(-1)

    def insert_completion(self, p_insert):
        final_text = self.start + p_insert + self.end
        self.set_edit_text(final_text)
        self.set_edit_pos(len(self.start) + len(p_insert))

    def _complete(self):
        pos = self.edit_pos
        text = self.edit_text

        word_before_cursor, is_first, start = _get_word_before_pos(text, pos)
        completions = self.completer.get_completions(word_before_cursor,
                                                     is_first)

        if not completions:
            return
        elif len(completions) > 1:  # TODO multiple completions
            replacement = commonprefix(completions)

            zero_candidate = replacement if replacement else word_before_cursor

            if zero_candidate != completions[0]:
                completions.insert(0, zero_candidate)

            # set slices before and after completion
            start, end = text[:start], text[pos:]
            self.start = start
            self.end = end

            urwid.emit_signal(self, 'show_completions', completions)
        else:
            replacement = completions[0]

        if replacement == word_before_cursor:
            return  # Don't complete what is already completed

        offset = len(replacement) - len(word_before_cursor)
        final_text = text[:pos] + replacement[-offset:] + text[pos:]
        self.set_edit_text(final_text)
        self.set_edit_pos(pos + offset)

    def keypress(self, p_size, p_key):
        dispatch = {
            'enter': self._emit_command,
            'esc': self._blur,
            'up': self._history_back,
            'down': self._history_next,
            'tab': self._complete
        }

        try:
            dispatch[p_key]()
        except KeyError:
            super().keypress(p_size, p_key)
